/*****************************************************************************/

/*
 *      config.c  --  USBD configuration files.
 *
 *      Copyright (C) 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "usbd.h"

/* ---------------------------------------------------------------------- */

LIST_HEAD(driverlist);

/* ---------------------------------------------------------------------- */

static void freedrvmatch(struct drivermatch *dm)
{
	free(dm);
}

static void freealldrvmatch(struct driver *drv)
{
	struct drivermatch *dm;

	while (!list_empty(&drv->matches)) {
		dm = list_entry(drv->matches.next, struct drivermatch, list);
		list_del(&dm->list);
		freedrvmatch(dm);
	}
}

static void freedrv(struct driver *drv)
{
	unloaddriver(drv, NULL);
	list_del(&drv->list);
	freealldrvmatch(drv);
	free(drv);
}

static struct driver *finddrv(unsigned int type, const char *name, unsigned int create)
{
	struct list_head *list;
	struct driver *drv;
	unsigned int len;

	for (list = driverlist.next; list != &driverlist; list = list->next) {
		drv = list_entry(list, struct driver, list);
		if (drv->type != type)
			continue;
		if (strcmp(drv->name, name))
			continue;
		return drv;
	}
	if (!create)
		return NULL;
	len = strlen(name);
	if (!(drv = malloc(sizeof(struct driver) + len + 1)))
		return NULL;
	INIT_LIST_HEAD(&drv->matches);
	drv->flags = 0;
	drv->type = type;
	drv->name = (char *)(drv + 1);
	strcpy((char *)drv->name, name);
	list_add_tail(&drv->list, &driverlist);
	return drv;
}

#if 0
struct driver *finddriver(unsigned int type, const char *name)
{
	return finddrv(type, name, 0);
}
#endif

static int parseline(char *line)
{
	struct drivermatch dm, *dmp;
	struct driver *drv;
	char *cp;
	struct stat statbuf;

	dm.vendorid = dm.productid = dm.class = dm.subclass = dm.protocol = -1;
	dm.priority = 0;
	while (isspace(*line))
		line++;
	if (*line == '#' || *line == ';')
		return 0;
	while ((cp = strtok(line, " \t"))) {
		line = NULL;
		if (!strcmp(cp, "vendor")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			dm.vendorid = strtoul(cp, NULL, 0);
			continue;
		}
		if (!strcmp(cp, "product")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			dm.productid = strtoul(cp, NULL, 0);
			continue;
		}
		if (!strcmp(cp, "class")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			dm.class = strtoul(cp, NULL, 0);
			continue;
		}
		if (!strcmp(cp, "subclass")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			dm.subclass = strtoul(cp, NULL, 0);
			continue;
		}
		if (!strcmp(cp, "protocol")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			dm.protocol = strtoul(cp, NULL, 0);
			continue;
		}
		if (!strcmp(cp, "priority") || !strcmp(cp, "prio")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			dm.priority = strtoul(cp, NULL, 0);
			continue;
		}
		if (!strcmp(cp, "kmod")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			if (!(drv = finddrv(DRVTYPE_KMOD, cp, 1)))
				return -1;
			if (!(dmp = malloc(sizeof(struct drivermatch))))
				return -1;
			*dmp = dm;
			list_add_tail(&dmp->list, &drv->matches);
			return 0;
		}
		if (!strcmp(cp, "user")) {
			if (!(cp = strtok(line, " \t")))
				return -1;
			if (stat(cp, &statbuf)) {
				lprintf(1, "usermode driver \"%s\" does not exist (%s, %d)\n", cp, strerror(errno), errno);
				return -1;
			}
			if (access(cp, X_OK)) {
				lprintf(1, "usermode driver \"%s\" is not executable (%s, %d)\n", cp, strerror(errno), errno);
				return -1;
			}
			if (!S_ISREG(statbuf.st_mode)) {
				lprintf(1, "usermode driver \"%s\" is not a regular file\n", cp);
				return -1;
			}
			if (!(drv = finddrv(DRVTYPE_USER, cp, 1)))
				return -1;
			if (!(dmp = malloc(sizeof(struct drivermatch))))
				return -1;
			*dmp = dm;
			list_add_tail(&dmp->list, &drv->matches);
			return 0;
		}
		if (!strcmp(cp, "builtin")) {
			if (!(cp = strtok(line, " \t")))
				cp = "?";
			if (!(drv = finddrv(DRVTYPE_BUILTIN, cp, 1)))
				return -1;
			if (!(dmp = malloc(sizeof(struct drivermatch))))
				return -1;
			*dmp = dm;
			list_add_tail(&dmp->list, &drv->matches);
			return 0;
		}
		lprintf(3, "unknown token \"%s\"\n", cp);
	}
	return 0;
}

static void parsefile(const char *name)
{
	int fd, i;
	char buf[16384];
	char *start, *end, *lineend;
	unsigned int linenr = 1;

	if ((fd = open(name, O_RDONLY)) == -1) {
		lprintf(1, "error opening config file \"%s\", %s (%d)\n", name, strerror(errno), errno);
		return;
	}
	start = buf;
	if ((i = read(fd, buf, sizeof(buf)-1)) == -1) {
		lprintf(1, "error reading config file \"%s\", %s (%d)\n", name, strerror(errno), errno);
		close(fd);
		return;
	}
	close(fd);
	/* parse lines */
	end = buf + i;
	*end = 0;
	while (start < end) {
		lineend = strchr(start, '\n');
		if (!lineend)
			break;
		*lineend = 0;
		if (parseline(start))
			lprintf(2, "Error in config file \"%s\" line %u\n", name, linenr);
		start = lineend + 1;
		linenr++;
	}
}

void parseconfig(void)
{
	struct list_head *list;
	struct driver *drv;
	DIR *d;
	struct dirent *de;
	char name[256];

	/* erase all existing matches */
	for (list = driverlist.next; list != &driverlist; list = list->next) {
		drv = list_entry(list, struct driver, list);
		freealldrvmatch(drv);
	}
	d = opendir(PACKAGE_CONFIG_DIR);
	if (!d) {
		lprintf(1, "Cannot open config files in \"" PACKAGE_CONFIG_DIR "\"\n");
		return;
	}
	while ((de = readdir(d))) {
		if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, ".."))
			continue;
		snprintf(name, sizeof(name), "%s/%s", PACKAGE_CONFIG_DIR, de->d_name);
		parsefile(name);
	}
	/* remove drivers with no attached matches records */
	for (list = driverlist.next; list != &driverlist; ) {
		drv = list_entry(list, struct driver, list);
		list = list->next;
		if (list_empty(&drv->matches))
			freedrv(drv);
	}
}

/* ---------------------------------------------------------------------- */

void dumpconfig(void)
{
	static const char *typelist[] = { "kmod", "user", "builtin" };
	struct list_head *list, *list2;
	struct driver *drv;
	struct drivermatch *dm;
	char m[128];
	char f[64];
	char *cp;

	for (list = driverlist.next; list != &driverlist; list = list->next) {
		drv = list_entry(list, struct driver, list);
		for (list2 = drv->matches.next; list2 != &drv->matches; list2 = list2->next) {
			dm = list_entry(list2, struct drivermatch, list);
			cp = m;
			if (dm->vendorid != -1)
				cp += sprintf(cp, "vendor 0x%04x ", dm->vendorid);
			if (dm->productid != -1)
				cp += sprintf(cp, "product 0x%04x ", dm->productid);
			if (dm->class != -1)
				cp += sprintf(cp, "class 0x%02x ", dm->class);
			if (dm->subclass != -1)
				cp += sprintf(cp, "subclass 0x%02x ", dm->subclass);
			if (dm->protocol != -1)
				cp += sprintf(cp, "protocol 0x%02x ", dm->protocol);
			if (dm->priority)
				cp += sprintf(cp, "priority %u ", dm->priority);
			*cp = 0;
			cp = f;
			if (drv->flags & DRVFLAG_NEEDED)
				cp += sprintf(cp, ",needed");
			if (drv->flags & DRVFLAG_LOADED)
				cp += sprintf(cp, ",loaded");
			*cp++ = ')';
			*cp = 0;
			f[0] = (cp <= f+2) ? 0 : '(';
			lprintf(1, "%s%s %s %s\n", m, (drv->type < 3) ? typelist[drv->type] : "??", drv->name, f);
		}
	}
}
