/*****************************************************************************/

/*
 *      example.c  --  USB usermode driver example.
 *
 *      Copyright (C) 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#ifdef HAVE_LINUX_USB_H
#include <linux/usb.h>
#else
#include "usb.h"
#endif

/* ---------------------------------------------------------------------- */

static unsigned int logging = 0;
static unsigned int verboselevel = 10;
static int fd = -1;

/* ---------------------------------------------------------------------- */

int lprintf(unsigned int vl, const char *format, ...)
{
        va_list ap;
        int r;

        if (vl > verboselevel)
                return 0;
        va_start(ap, format);
#ifdef HAVE_VSYSLOG
        if (logging) {
                static const int logprio[] = { LOG_CRIT, LOG_ERR, LOG_INFO };
                vsyslog((vl > 2) ? LOG_DEBUG : logprio[vl], format, ap);
                r = 0;
        } else
#endif
                r = vfprintf(stderr, format, ap);
        va_end(ap);
	if (!vl)
		exit(1);
        return r;
}

/* ---------------------------------------------------------------------- */

static int usb_control_msg(int fd, u_int8_t requesttype, u_int8_t request, u_int16_t value,
                           u_int16_t index, unsigned int size, void *data)
{
        struct usb_proc_ctrltransfer ctrl;

        ctrl.requesttype = requesttype;
        ctrl.request = request;
        ctrl.value = value;
        ctrl.index = index;
        ctrl.length = size;
        ctrl.data = data;
        return ioctl(fd, USB_PROC_CONTROL, &ctrl);
}

static int usb_getconfigdesc(unsigned char *buf, unsigned int len, unsigned int nr)
{
        unsigned int sz;
	
        if (usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_CONFIG << 8) | nr,
			    0, USB_DT_CONFIG_SIZE, buf) < 0)
                return -1;
        if (buf[0] < USB_DT_CONFIG_SIZE || buf[1] != USB_DT_CONFIG)
		return -1;
        sz = buf[2] | buf[3] << 8;
        if (sz > len)
		sz = len;
        if (usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_CONFIG << 8) | nr,
                            0, sz, buf) < 0)
                return -1;
        return sz;
}

/* ---------------------------------------------------------------------- */


static void usbconnect(unsigned int busnum, unsigned int devnum, const char *procpath,
		       unsigned int vendorid, unsigned int productid)
{
	unsigned char buf[1024];
	char buf2[256];
	char *cp;
	int i, j;

	lprintf(1, "USB connect: bus %2d device %3d vendor 0x%04x product 0x%04x path %s\n",
		busnum, devnum, vendorid, productid, procpath);
	fd = open(procpath, O_RDWR);
	if (fd == -1)
		lprintf(0, "cannot open \"%s\", %s (%d)\n", procpath, strerror(errno), errno);
	i = usb_getconfigdesc(buf, sizeof(buf), 0);
	if (i <= 0)
		lprintf(0, "error fetching config0 descriptor\n");
	lprintf(1, "Config0 descriptor\n");
	cp = buf2;
	for (j = 0; j < i; j++) {
		sprintf(cp, "%02X ", buf[j]);
		cp += 3;
		if ((~j) & 15)
			continue;
		cp[-1] = 0;
		lprintf(1, "%04x: %s\n", j & ~15, buf2);
		cp = buf2;
	}
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
        static const struct option long_options[] = {
                { "usbd", 0, 0, 0x100 },
                { "syslog", 0, 0, 0x101 },
                 { 0, 0, 0, 0}
        };
	char *pbusnum, *pdevnum, *pprocpath, *pvendorid, *pproductid;
        int c;
        int err = 0;
	int usbdcalling = 0;

        /* parse arguments */
        while ((c = getopt_long(argc, argv, "", long_options, NULL)) != EOF)
                switch (c) {
#if defined(HAVE_SYSLOG) && defined(HAVE_VSYSLOG)
                case 0x101:
                        logging = 1;
                        break;
#endif

                case 0x100:
                        usbdcalling = 1;
			break;

                default:
                        err++;
                        break;
                }
        if (err) {
                fprintf(stderr, "usage: example [--usbd] [--syslog]\n");
                exit(1);
        }
#if defined(HAVE_SYSLOG) && defined(HAVE_VSYSLOG)
	if (logging)
		openlog("example", LOG_PID, LOG_DAEMON);
#endif
	if (!usbdcalling)
		lprintf(0, "example must be called by usbd\n");
	if (!(pbusnum = getenv("USBD_BUSNUM")) ||
	    !(pdevnum = getenv("USBD_DEVNUM")) ||
	    !(pprocpath = getenv("USBD_PROCPATH")) ||
	    !(pvendorid = getenv("USBD_VENDORID")) ||
	    !(pproductid = getenv("USBD_PRODUCTID")))
		lprintf(0, "Invalid environment\n");
	usbconnect(strtoul(pbusnum, NULL, 0), strtoul(pdevnum, NULL, 0), pprocpath, 
		   strtoul(pvendorid, NULL, 0), strtoul(pproductid, NULL, 0));
	exit(0);
}
