/*****************************************************************************/

/*
 *      init.c  --  USB daemon initialization routines.
 *
 *      Copyright (C) 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include "usbd.h"

/* ---------------------------------------------------------------------- */

#define PIDFILE PACKAGE_LOCALSTATE_DIR "/run/usbd.pid"

/* --------------------------------------------------------------------- */

static void write_pid(pid_t pid)
{
        FILE *f = fopen(PIDFILE, "a");
        if (!f) {
                lprintf(1, "cannot open PID file %s\n", PIDFILE);
                return;
        }
        fprintf(f, "%ld\n", (long)pid);
        fclose(f);
}

static void kill_usbd(void)
{
        FILE *f = fopen(PIDFILE, "r");
	long lpid;

        if (!f) {
                lprintf(1, "cannot open PID file %s\n", PIDFILE);
                return;
        }
        while (fscanf(f, " %ld \n", &lpid) == 1) 
                if (lpid <= 0 || lpid == getpid())
                        lprintf(1, "invalid pid %ld\n", lpid);
                else {
                        kill(lpid, SIGTERM);
                        lprintf(1, "terminating process %ld\n", lpid);
                }
        fclose(f);
        unlink(PIDFILE);
}

/* ---------------------------------------------------------------------- */


int usbdinit(int argc, char *argv[])
{
        static const struct option long_options[] = {
                { "log", 0, 0, 'l' },
                { "kill", 0, 0, 'k' },
                { "verbose", 1, 0, 0x100 },
                { "no-daemon", 0, 0, 'n' },
                { 0, 0, 0, 0}
        };
        int c;
        int err = 0;
        pid_t pid;
        int daemon = 1;
	int fd;

        printf(PACKAGE " v" VERSION " (c) 1999 by Thomas Sailer\n");
        /* parse arguments */
        while ((c = getopt_long(argc, argv, "lknv", long_options, NULL)) != EOF)
                switch (c) {
#if defined(HAVE_SYSLOG) && defined(HAVE_VSYSLOG)
                case 'l':
                        logging = 1;
                        break;
#endif

                case 'k':
                        kill_usbd();
                        exit(0);

                case 'n':
                        daemon = 0;
                        break;

		case 0x100:
			verboselevel = strtoul(optarg, NULL, 0);
			break;

                case 'v':
                        verboselevel++;
                        break;

                default:
                        err++;
                        break;
                }
        if (err) {
                fprintf(stderr, "usage: usbd [-l] [-k] [-n] [-v]\n");
                exit(1);
        }
	/* init */
	parseconfig();
	dumpconfig();
	fd = open(PACKAGE_PROCBUSUSB_DIR "/devices", O_RDONLY);
	if (fd == -1) {
		request_module("usbcore", 0);
		fd = open("/proc/bus/usb/devices", O_RDONLY);
	}
	if (fd == -1)
		lprintf(0, "cannot open \"" PACKAGE_PROCBUSUSB_DIR "/devices \": %s (%d)\n", strerror(errno), errno);
	/* fork if daemon mode */
        if (!daemon) {
                setsid(); /* become a process group leader and drop controlling terminal */
                fclose(stdin); /* no more standard in */
#if defined(HAVE_SYSLOG) && defined(HAVE_VSYSLOG)
                if (logging)
                        openlog("usbd", LOG_PID, LOG_DAEMON);
#endif
                return fd;
        }
        switch (pid = fork()) {
        case -1:
                lprintf(0, "fork: %s (%d)\n", strerror(errno), errno);
                exit(1);

        case 0: /* child process */
                setsid(); /* become a process group leader and drop controlling terminal */
                fclose(stdin); /* no more standard in */
#if defined(HAVE_SYSLOG) && defined(HAVE_VSYSLOG)
                if (logging)
                        openlog("usbd", LOG_PID, LOG_DAEMON);
#endif
                return fd;

        default:
                close(fd);
                write_pid(pid);
                lprintf(3, "child pid: %ld\n", (long)pid);
                break;
        }
        exit(0);
}
