/*****************************************************************************/

/*
 *      loader.c  --  USBD loading/forking kernel/user drivers.
 *
 *      Copyright (C) 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _USE_BSD
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "usbd.h"

/* ---------------------------------------------------------------------- */

struct pidlist {
	struct list_head list;
	pid_t pid;
};

static LIST_HEAD(pidlist);

/* ---------------------------------------------------------------------- */

static int callmodprobe(char * const *argv)
{
	static char * const modprobe_envp[] = { "HOME=/", "TERM=linux", "PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL };
	pid_t pid;
	int status;

	pid = fork();
	switch (pid) {
	case 0: /* child */
		setsid();
		execve(UTIL_MODPROBE, argv, modprobe_envp);
		lprintf(1, "exec: %s (%u)\n", strerror(errno), errno);
		return -1;

	case -1: /* error */
		lprintf(1, "fork: %s (%u)\n", strerror(errno), errno);
		return -1;

	default: /* parent */
		if (wait4(pid, &status, 0, NULL) == -1) {
			lprintf(1, "wait4: %s (%u)\n", strerror(errno), errno);
			return -1;
		}
		if (status)
			lprintf(1, UTIL_MODPROBE " returned %d\n", status);
		return status;
	}
}

int request_module(const char *modulename, int permanent)
{
	char *argv[5];
	char **ap = argv;

	*ap++ = UTIL_MODPROBE;
	if (logging)
		*ap++ = "-s";
	if (!permanent)
		*ap++ = "-k";
	*ap++ = modulename;
	*ap++ = NULL;
	return callmodprobe(argv);
}

int remove_module(const char *modulename)
{
	char *argv[5];
	char **ap = argv;

	*ap++ = UTIL_MODPROBE;
	if (logging)
		*ap++ = "-s";
	*ap++ = "-r";
	*ap++ = modulename;
	*ap++ = NULL;
	return callmodprobe(argv);
}

/* ---------------------------------------------------------------------- */

static void childdeath(int sig, siginfo_t *si, void *ptr)
{
	pid_t pid;
	int status;
	struct list_head *list;
	struct pidlist *pl;

	while ((pid = wait4(-1, &status, WNOHANG, NULL)) != -1) {
		lprintf(2, "child %ld died, code %d\n", (long)pid, si->si_code);
		for (list = pidlist.next; list != &pidlist; list = list->next) {
			pl = list_entry(list, struct pidlist, list);
			if (pl->pid != pid)
				continue;
			list_del(&pl->list);
			free(pl);
			goto found;
		}
		lprintf(1, "warning: child pid %ld died, but not in child list\n", (long)pid);
	found:
	}
}

/* ---------------------------------------------------------------------- */

void initloader(void)
{
	struct sigaction sigact;

	sigact.sa_sigaction = childdeath;
	sigact.sa_flags = SA_SIGINFO | SA_RESTART | SA_NOCLDSTOP;
	sigemptyset(&sigact.sa_mask);
	sigaction(SIGCHLD, &sigact, NULL);
}

/* ---------------------------------------------------------------------- */

static int forkuser(struct driver *drv, struct usbdevnode *dev)
{
	char envbusnum[18];
	char envdevnum[18];
	char envprocpath[128];
	char envvendorid[22];
	char envproductid[24];
	char *argv[] = { drv->name, "--usbd", logging ? "--syslog" : NULL, NULL };
	char *envp[] = { "HOME=/", "TERM=linux", "PATH=/sbin:/usr/sbin:/bin:/usr/bin", envbusnum, envdevnum, envprocpath, envvendorid, envproductid };
	pid_t pid;

	if (!dev)
		return -1;
	
        switch (pid = fork()) {
        case -1:
                lprintf(1, "fork: %s (%d)\n", strerror(errno), errno);
                return -1;

        case 0: /* child process */
                setsid(); /* become a process group leader and drop controlling terminal */
                close(0); /* no more standard in */
		snprintf(envbusnum, sizeof(envbusnum), "USBD_BUSNUM=%d", dev->bus->busnum);
		snprintf(envdevnum, sizeof(envdevnum), "USBD_DEVNUM=%d", dev->devnum);
		snprintf(envprocpath, sizeof(envprocpath), "USBD_PROCPATH=" PACKAGE_PROCBUSUSB_DIR "/%03d/%03d", dev->bus->busnum, dev->devnum);
		snprintf(envvendorid, sizeof(envvendorid), "USBD_VENDORID=0x%04x", dev->vendorid);
		snprintf(envproductid, sizeof(envproductid), "USBD_PRODUCTID=0x%04x", dev->productid);
		execve(drv->name, argv, envp);
		lprintf(0, "execve: %s, %s (%d)\n", strerror(errno), errno);

        default:
		return 0;
        }
 }

void loaddriver(struct driver *drv, struct usbdevnode *dev)
{
	int i;

	if (drv->flags & DRVFLAG_LOADED)
		return;
	lprintf(3, "loading driver type %d name %s\n", drv->type, drv->name);
	switch (drv->type) {
	case DRVTYPE_KMOD:
		if ((i = request_module(drv->name, 1))) {
			lprintf(1, "error %d loading kernel module \"%s\"\n", i, drv->name);
			return;
		}
		drv->flags |= DRVFLAG_LOADED;
		return;

	case DRVTYPE_USER:
		if (forkuser(drv, dev))
			break;
		drv->flags &= ~(DRVFLAG_LOADED | DRVFLAG_NEEDED);
		break;

	case DRVTYPE_BUILTIN:
		drv->flags |= DRVFLAG_LOADED;
		return;

	}
}

void unloaddriver(struct driver *drv, struct usbdevnode *dev)
{
	int i;

	if (!(drv->flags & DRVFLAG_LOADED))
		return;
	lprintf(3, "unloading driver type %d name %s\n", drv->type, drv->name);
	switch (drv->type) {
	case DRVTYPE_KMOD:
		if ((i = remove_module(drv->name))) {
			lprintf(1, "error %d unloading kernel module \"%s\"\n", i, drv->name);
			return;
		}
		drv->flags &= ~DRVFLAG_LOADED;
		return;

	case DRVTYPE_USER:
		drv->flags &= ~(DRVFLAG_LOADED | DRVFLAG_NEEDED);
		break;

	case DRVTYPE_BUILTIN:
		drv->flags &= ~DRVFLAG_LOADED;
		return;

	}
}

/* ---------------------------------------------------------------------- */

static struct driver *findbestdriver(struct usbdevnode *dev, unsigned int cls)
{
	struct list_head *list, *list2;
	struct driver *drv;
	struct drivermatch *dm;
	struct driver *drv2 = NULL;
	unsigned int prio = 0;

	if (cls >= dev->numclasses)
		return NULL;
	for (list = driverlist.next; list != &driverlist; list = list->next) {
		drv = list_entry(list, struct driver, list);
		for (list2 = drv->matches.next; list2 != &drv->matches; list2 = list2->next) {
			dm = list_entry(list2, struct drivermatch, list);
			if (dm->vendorid != -1 && dm->vendorid != dev->vendorid)
				continue;
			if (dm->productid != -1 && dm->productid != dev->productid)
				continue;
			if (dm->class != -1 && dm->class != dev->class[cls].class)
				continue;
			if (dm->subclass != -1 && dm->subclass != dev->class[cls].subclass)
				continue;
			if (dm->protocol != -1 && dm->protocol != dev->class[cls].protocol)
				continue;
			if (dm->priority >= prio) {
				prio = dm->priority;
				drv2 = drv;
			}
		}
	}
	return drv2;
}

static void markdev(struct usbdevnode *dev, unsigned int markall)
{
	struct driver *drv;
	unsigned int cls;

	for (cls = 0; cls < dev->numclasses; cls++) {
		drv = findbestdriver(dev, cls);
		if (drv) {
			if (markall || drv->type != DRVTYPE_USER)
				drv->flags |= DRVFLAG_NEEDED;
		} else if (dev->class[cls].class != 0) {
			lprintf(1, "do not know how to handle bus %d device %d if %d class 0x%02x\n", 
				dev->devnum, dev->bus->busnum, cls, dev->class[cls].class);
		}
	}
}

static void walkdevlist(struct list_head *list)
{
	struct list_head *list2;
	struct driver *drv;
	struct usbdevnode *dev;

	for (list2 = list->next; list2 != list; list2 = list2->next) {
		dev = list_entry(list2, struct usbdevnode, list);
		markdev(dev, 0);
		walkdevlist(&dev->childlist);
	}
}

static void calcneeded(void)
{
	struct list_head *list;
	struct driver *drv;
	struct usbbusnode *bus;

	/* mark all as not needed */
	for (list = driverlist.next; list != &driverlist; list = list->next) {
		drv = list_entry(list, struct driver, list);
		drv->flags &= ~DRVFLAG_NEEDED;
	}
	/* find out which drivers are needed */
	for (list = usbbuslist.next; list != &usbbuslist; list = list->next) {
		bus = list_entry(list, struct usbbusnode, list);
		walkdevlist(&bus->childlist);
	}
dumpconfig();
}

static void doload(struct usbdevnode *dev)
{
	struct list_head *list;
	struct driver *drv;

	/* load/unload drivers */
	for (list = driverlist.next; list != &driverlist; list = list->next) {
		drv = list_entry(list, struct driver, list);
		if ((drv->flags & DRVFLAG_NEEDED) && !(drv->flags & DRVFLAG_LOADED))
			loaddriver(drv, dev);
		else if (!(drv->flags & DRVFLAG_NEEDED) && (drv->flags & DRVFLAG_LOADED))
			unloaddriver(drv, dev);
	}
}

/* ---------------------------------------------------------------------- */

void devtree_busconnect(struct usbbusnode *bus)
{
	lprintf(3, "Bus Connect: busnum %d\n", bus->busnum);
	calcneeded();
	doload(NULL);

	devtree_dump();
}

void devtree_busdisconnect(struct usbbusnode *bus)
{
	lprintf(3, "Bus Disconnect: busnum %d\n", bus->busnum);
	calcneeded();
	doload(NULL);

	devtree_dump();
}

void devtree_devconnect(struct usbdevnode *dev)
{
	lprintf(3, "Device Connect: busnum %d devnum %d parent %d vendor %04x product %04x\n",
		dev->bus->busnum, dev->devnum, dev->parent ? dev->parent->devnum : -1, dev->vendorid, dev->productid);
	calcneeded();
	markdev(dev, 1);
	doload(dev);

	devtree_dump();
}

void devtree_devdisconnect(struct usbdevnode *dev)
{
	lprintf(3, "Device Disconnect: busnum %d devnum %d parent %d vendor %04x product %04x\n",
		dev->bus->busnum, dev->devnum, dev->parent ? dev->parent->devnum : -1, dev->vendorid, dev->productid);
	calcneeded();
	doload(NULL);

	devtree_dump();
}

