/*****************************************************************************/

/*
 *      usbd.c  --  USB daemon skeleton.
 *
 *      Copyright (C) Thomas Sailer, sailer@ife.ee.ethz.ch
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#include <sys/types.h>
#include <sys/poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#include "usbd.h"

/* ---------------------------------------------------------------------- */

unsigned int verboselevel = 0;
unsigned int logging = 0;

/* ---------------------------------------------------------------------- */

int lprintf(unsigned int vl, const char *format, ...)
{
        va_list ap;
        int r;

        if (vl > verboselevel)
                return 0;
        va_start(ap, format);
#ifdef HAVE_VSYSLOG
        if (logging) {
                static const int logprio[] = { LOG_CRIT, LOG_ERR, LOG_INFO };
                vsyslog((vl > 2) ? LOG_DEBUG : logprio[vl], format, ap);
                r = 0;
        } else
#endif
                r = vfprintf(stderr, format, ap);
        va_end(ap);
	if (!vl)
		exit(1);
        return r;
}

/* ---------------------------------------------------------------------- */

static void sighup(int dummy)
{	
	parseconfig();
}

static void sigusr1(int dummy)
{
	devtree_dump();
}

static void sigusr2(int dummy)
{
	dumpconfig();
}

static void setupsigs(void)
{
	struct sigaction sa;

	sa.sa_handler = sighup;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sigaction(SIGHUP, &sa, NULL);
	sa.sa_handler = sigusr1;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sigaction(SIGUSR1, &sa, NULL);
	sa.sa_handler = sigusr2;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sigaction(SIGUSR2, &sa, NULL);
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	struct pollfd pfd;
	sigset_t sigmask;
	int err;

	pfd.fd = usbdinit(argc, argv);
	pfd.events = POLLIN;
	/* set up signal handlers */
	setupsigs();
	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGHUP);
	sigaddset(&sigmask, SIGUSR1);
	sigaddset(&sigmask, SIGUSR2);
	sigprocmask(SIG_BLOCK, &sigmask, NULL);
	devtree_parsedevfile(pfd.fd);
	devtree_processchanges();
	for (;;) {
		sigprocmask(SIG_UNBLOCK, &sigmask, NULL);
		do {
			err = poll(&pfd, 1, 60000);
		} while (err == -1 && errno == EINTR);
		sigprocmask(SIG_BLOCK, &sigmask, NULL);
		if (err == -1)
			lprintf(0, "poll: %s (%d)\n", strerror(errno), errno);
		if (err > 0)
			lprintf(3, "Connect/Disconnect event\n");
		devtree_parsedevfile(pfd.fd);
		devtree_processchanges();
	}
}
